import datetime
from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, TypeVar, Union
from uuid import UUID

from attrs import define as _attrs_define
from attrs import field as _attrs_field
from dateutil.parser import isoparse

from ..models.downlink_message_status import DownlinkMessageStatus
from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.rpc_req import RpcReq
    from ..models.rpc_rsp import RpcRsp


T = TypeVar("T", bound="DownlinkMessage")


@_attrs_define
class DownlinkMessage:
    """
    Attributes:
        id (UUID): The ID of the downlink message Example: 7527bf1c-9868-4afd-b07d-16dc7eb7bed3.
        created_at (datetime.datetime): The time the downlink message was created
        updated_at (datetime.datetime): The time the downlink message was last updated
        device_id (UUID): The ID of the device the message is for Example: 5f4b1b2b-3b4d-4b5e-8c6f-7d8e9f0a1b2c.
        payload_type (int): The type of payload
        auth (int): The auth level of the message
        rpc_req (RpcReq):
        status (DownlinkMessageStatus): Status of downlink message
        rpc_rsp (Union[Unset, RpcRsp]):
        send_wait_timeout_ms (Union[Unset, int]): Maximum time to wait (in milliseconds) for the device to send a packet
            before expiring. If 0 or not set, the RPC was sent immediately using the device's last route.
        sent_at (Union[Unset, datetime.datetime]): The time the downlink message was sent
        expires_at (Union[Unset, datetime.datetime]): The time the downlink message expires
        completed_at (Union[Unset, datetime.datetime]): The time the downlink message was completed
    """

    id: UUID
    created_at: datetime.datetime
    updated_at: datetime.datetime
    device_id: UUID
    payload_type: int
    auth: int
    rpc_req: "RpcReq"
    status: DownlinkMessageStatus
    rpc_rsp: Union[Unset, "RpcRsp"] = UNSET
    send_wait_timeout_ms: Unset | int = UNSET
    sent_at: Unset | datetime.datetime = UNSET
    expires_at: Unset | datetime.datetime = UNSET
    completed_at: Unset | datetime.datetime = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        id = str(self.id)

        created_at = self.created_at.isoformat()

        updated_at = self.updated_at.isoformat()

        device_id = str(self.device_id)

        payload_type = self.payload_type

        auth = self.auth

        rpc_req = self.rpc_req.to_dict()

        status = self.status.value

        rpc_rsp: Unset | dict[str, Any] = UNSET
        if not isinstance(self.rpc_rsp, Unset):
            rpc_rsp = self.rpc_rsp.to_dict()

        send_wait_timeout_ms = self.send_wait_timeout_ms

        sent_at: Unset | str = UNSET
        if not isinstance(self.sent_at, Unset):
            sent_at = self.sent_at.isoformat()

        expires_at: Unset | str = UNSET
        if not isinstance(self.expires_at, Unset):
            expires_at = self.expires_at.isoformat()

        completed_at: Unset | str = UNSET
        if not isinstance(self.completed_at, Unset):
            completed_at = self.completed_at.isoformat()

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "id": id,
                "createdAt": created_at,
                "updatedAt": updated_at,
                "deviceId": device_id,
                "payloadType": payload_type,
                "auth": auth,
                "rpcReq": rpc_req,
                "status": status,
            }
        )
        if rpc_rsp is not UNSET:
            field_dict["rpcRsp"] = rpc_rsp
        if send_wait_timeout_ms is not UNSET:
            field_dict["sendWaitTimeoutMs"] = send_wait_timeout_ms
        if sent_at is not UNSET:
            field_dict["sentAt"] = sent_at
        if expires_at is not UNSET:
            field_dict["expiresAt"] = expires_at
        if completed_at is not UNSET:
            field_dict["completedAt"] = completed_at

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.rpc_req import RpcReq
        from ..models.rpc_rsp import RpcRsp

        d = dict(src_dict)
        id = UUID(d.pop("id"))

        created_at = isoparse(d.pop("createdAt"))

        updated_at = isoparse(d.pop("updatedAt"))

        device_id = UUID(d.pop("deviceId"))

        payload_type = d.pop("payloadType")

        auth = d.pop("auth")

        rpc_req = RpcReq.from_dict(d.pop("rpcReq"))

        status = DownlinkMessageStatus(d.pop("status"))

        _rpc_rsp = d.pop("rpcRsp", UNSET)
        rpc_rsp: Unset | RpcRsp
        if isinstance(_rpc_rsp, Unset):
            rpc_rsp = UNSET
        else:
            rpc_rsp = RpcRsp.from_dict(_rpc_rsp)

        send_wait_timeout_ms = d.pop("sendWaitTimeoutMs", UNSET)

        _sent_at = d.pop("sentAt", UNSET)
        sent_at: Unset | datetime.datetime
        if isinstance(_sent_at, Unset):
            sent_at = UNSET
        else:
            sent_at = isoparse(_sent_at)

        _expires_at = d.pop("expiresAt", UNSET)
        expires_at: Unset | datetime.datetime
        if isinstance(_expires_at, Unset):
            expires_at = UNSET
        else:
            expires_at = isoparse(_expires_at)

        _completed_at = d.pop("completedAt", UNSET)
        completed_at: Unset | datetime.datetime
        if isinstance(_completed_at, Unset):
            completed_at = UNSET
        else:
            completed_at = isoparse(_completed_at)

        downlink_message = cls(
            id=id,
            created_at=created_at,
            updated_at=updated_at,
            device_id=device_id,
            payload_type=payload_type,
            auth=auth,
            rpc_req=rpc_req,
            status=status,
            rpc_rsp=rpc_rsp,
            send_wait_timeout_ms=send_wait_timeout_ms,
            sent_at=sent_at,
            expires_at=expires_at,
            completed_at=completed_at,
        )

        downlink_message.additional_properties = d
        return downlink_message

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



#!/usr/bin/env python3

import ctypes

from infuse_iot.commands import InfuseRpcCommand
from infuse_iot.util.ctypes import VLACompatLittleEndianStruct, bytes_to_uint8
from infuse_iot.zephyr.errno import errno


class generic(InfuseRpcCommand):
    NAME = "generic"
    HELP = "Generic RPC sender"
    DESCRIPTION = "Generic RPC sender"
    COMMAND_ID = -1

    class response(VLACompatLittleEndianStruct):
        _fields_ = []
        vla_field = ("val", 0 * ctypes.c_uint8)
        _pack_ = 1

    @classmethod
    def add_parser(cls, parser):
        parser.add_argument("--rpc", type=int, required=True, help="RPC command ID")
        parser.add_argument("--args", type=str, default="", help="Arguments as byte string")

    def __init__(self, args):
        self.COMMAND_ID = args.rpc
        self._arg_bytes = bytes.fromhex(args.args)

    def request_struct(self):
        class request(ctypes.LittleEndianStructure):
            _fields_ = [
                ("val", len(self._arg_bytes) * ctypes.c_uint8),
            ]
            _pack_ = 1

        return request(bytes_to_uint8(self._arg_bytes))

    def handle_response(self, return_code, response):
        print(f"Return Code: {return_code} ({errno.strerror(-return_code)})")
        if len(response.val) > 0:
            print(f"   Response: {bytes(response.val).hex()}")
        else:
            print("   Response: None")



#!/usr/bin/env python3

"""Synchronise data logger state from remote devices"""

__author__ = "Jordan Yates"
__copyright__ = "Copyright 2025, Embeint Holdings Pty Ltd"

import binascii
import glob
import os
import pathlib

from rich.live import Live
from rich.progress import (
    DownloadColumn,
    Progress,
    TransferSpeedColumn,
)
from rich.status import Status
from rich.table import Table

from infuse_iot.commands import InfuseCommand
from infuse_iot.definitions.rpc import data_logger_read, rpc_enum_data_logger
from infuse_iot.epacket.packet import Auth
from infuse_iot.generated.tdf_definitions import readings
from infuse_iot.rpc_client import RpcClient
from infuse_iot.socket_comms import (
    GatewayRequestConnectionRequest,
    LocalClient,
    default_multicast_address,
)
from infuse_iot.util.argparse import ValidDir


class DeviceState:
    BLOCK_SIZE = 512

    def __init__(self, path: pathlib.Path):
        self.path = path
        self.on_disk: int = 0
        self.on_device: int | None = None
        self.downloaded: int = 0
        if not self.path.exists():
            self.path.touch()
        else:
            self.on_disk = os.path.getsize(self.path) // self.BLOCK_SIZE

    def observe(self, announce: readings.announce | readings.announce_v2):
        self.on_device = announce.blocks

    def append_data(self, data: bytes):
        assert len(data) % self.BLOCK_SIZE == 0
        new_blocks = len(data) // self.BLOCK_SIZE

        with self.path.open("+ba") as f:
            f.write(data)
            self.on_disk += new_blocks
            self.downloaded += new_blocks


class SubCommand(InfuseCommand):
    NAME = "data_logger_sync"
    HELP = "Synchronise data logger state from remote devices"
    DESCRIPTION = "Synchronise data logger state from remote devices"

    def __init__(self, args):
        self._client = LocalClient(default_multicast_address(), 1.0)
        self._min_rssi: int | None = args.rssi
        self._app = args.app
        self._out = args.out
        self._blocks_max = args.blocks
        self._logger = args.logger
        self._device_state: dict[int, DeviceState] = {}

        self.bytes_to_read = 0
        self.pending_bytes = b""
        self.task = None
        self.state = "Scanning"
        self.progress = Progress(
            *Progress.get_default_columns(),
            DownloadColumn(),
            TransferSpeedColumn(),
        )

        for file in glob.glob(f"{self._out}/*_{self._logger}.bin"):
            file_path = pathlib.Path(file)
            name_parts = file_path.name.split("_")
            device_id = int(name_parts[0], 16)

            self._device_state[device_id] = DeviceState(file_path)

    @classmethod
    def add_parser(cls, parser):
        parser.add_argument("--out", "-o", type=ValidDir, required=True, help="Output folder for synced data")
        parser.add_argument("--blocks", "-b", type=int, default=500, help="Number of blocks to download per connection")
        parser.add_argument("--app", "-a", type=lambda x: int(x, 0), help="Application ID to limit sync to")
        parser.add_argument("--rssi", "-r", type=int, help="Minimum RSSI to attempt downloading data")
        logger_group = parser.add_mutually_exclusive_group(required=True)
        logger_group.add_argument(
            "--onboard",
            dest="logger",
            action="store_const",
            const=rpc_enum_data_logger.FLASH_ONBOARD,
            help="Synchronise onboard loggers",
        )
        logger_group.add_argument(
            "--removable",
            dest="logger",
            action="store_const",
            const=rpc_enum_data_logger.FLASH_REMOVABLE,
            help="Synchronise removable loggers",
        )

    def progress_table(self):
        table = Table()
        table.add_column("Device ID")
        table.add_column("On Disk", justify="right")
        table.add_column("On Device", justify="right")
        table.add_column("Downloaded", justify="right")
        for device, state in self._device_state.items():
            on_device = str(state.on_device) if state.on_device is not None else "?"
            percent = f"{100 * state.on_disk / state.on_device:.0f}" if state.on_device is not None else "?"
            table.add_row(f"{device:016x}", f"{state.on_disk} ({percent:>3s}%)", on_device, str(state.downloaded))

        meta = Table(box=None)
        meta.add_column()
        meta.add_row(table)
        meta.add_row(Status(self.state))
        meta.add_row(self.progress)

        return meta

    def state_update(self, live: Live, state: str):
        self.state = state
        live.update(self.progress_table())

    def data_progress_cb(self, offset: int, payload: bytes):
        if self.task is None:
            self.state = "Reading data logger"
            self.task = self.progress.add_task("", total=self.bytes_to_read)
        self.pending_bytes += payload
        self.progress.update(self.task, completed=offset)

    def handle_sync(self, live: Live, device_id: int, state: DeviceState):
        self.state_update(live, f"Connecting to {device_id:016X}")
        assert state.on_device is not None
        try:
            with self._client.connection(device_id, GatewayRequestConnectionRequest.DataType.COMMAND) as mtu:
                self.state_update(live, f"Downloading blocks from {device_id:016X}")
                rpc_client = RpcClient(self._client, mtu, device_id)
                blocks_pending = state.on_device - state.on_disk
                blocks_to_read = min(blocks_pending, self._blocks_max)
                last_block = state.on_disk + blocks_to_read - 1
                self.bytes_to_read = 512 * blocks_to_read
                params = data_logger_read.request(self._logger, state.on_disk, last_block)
                self.pending_bytes = b""
                hdr, rsp = rpc_client.run_data_recv_cmd(
                    data_logger_read.COMMAND_ID,
                    Auth.DEVICE,
                    bytes(params),
                    self.bytes_to_read,
                    self.data_progress_cb,
                    data_logger_read.response.from_buffer_copy,
                )
                if hdr.return_code == 0:
                    assert isinstance(rsp, data_logger_read.response)
                    if rsp.sent_len == len(self.pending_bytes) and rsp.sent_crc == binascii.crc32(self.pending_bytes):
                        state.append_data(self.pending_bytes)
                if self.task is not None:
                    self.progress.remove_task(self.task)
                    self.task = None

        except ConnectionRefusedError:
            self.state_update(live, "Scanning")
        except ConnectionAbortedError:
            self.state_update(live, "Scanning")

    def run(self):
        with Live(self.progress_table(), refresh_per_second=4) as live:
            for source, announce in self._client.observe_announce():
                self.state_update(live, "Scanning")
                if self._app and announce.application != self._app:
                    continue
                # Don't consider announce packets that don't contain information about the requested logger
                req_removable = self._logger == rpc_enum_data_logger.FLASH_REMOVABLE
                announce_is_removable = announce.flags & 0x01
                if req_removable != announce_is_removable:
                    continue

                if source.infuse_id not in self._device_state:
                    output_file = self._out / f"{source.infuse_id:016x}_{self._logger}.bin"
                    self._device_state[source.infuse_id] = DeviceState(output_file)
                state = self._device_state[source.infuse_id]
                state.observe(announce)

                # Is signal strong enough to connect?
                if self._min_rssi and source.rssi < self._min_rssi:
                    continue

                assert state.on_device is not None
                if state.on_disk < state.on_device:
                    self.handle_sync(live, source.infuse_id, state)



# Copyright (c) 2025 Embeint Holdings Pty Ltd
#
# SPDX-License-Identifier: Apache-2.0

import json
import subprocess
import sys

from infuse_iot.util.soc.soc import ProvisioningInterface


class NRFFamily:
    FICR_ADDRESS: int
    DEVICE_ID_OFFSET: int
    CUSTOMER_OFFSET: int

    @staticmethod
    def soc(device_info):
        raise NotImplementedError


class nRF52(NRFFamily):
    FICR_ADDRESS = 0x10000000
    DEVICE_ID_OFFSET = 0x60
    CUSTOMER_OFFSET = 0x80

    @staticmethod
    def soc(device_info):
        version: str = device_info["jlink"]["deviceVersion"]
        if version.startswith("NRF52840"):
            return "nRF52840"
        elif version.startswith("NRF52833"):
            return "nRF52833"
        else:
            raise NotImplementedError(f"Unhandled device {version}")


class nRF53(NRFFamily):
    FICR_ADDRESS = 0x00FF0000
    DEVICE_ID_OFFSET = 0x204
    CUSTOMER_OFFSET = 0x100

    @staticmethod
    def soc(device_info):
        version: str = device_info["jlink"]["deviceVersion"]
        if version.startswith("NRF5340"):
            return "nRF5340"
        else:
            raise NotImplementedError(f"Unhandled device {version}")


class nRF54L(NRFFamily):
    FICR_ADDRESS = 0x00FFC000
    DEVICE_ID_OFFSET = 0x304
    CUSTOMER_OFFSET = 0x500

    @staticmethod
    def soc(device_info):
        version: str = device_info["jlink"]["deviceVersion"]
        if version.startswith("NRF54L15"):
            return "nRF54L15"
        elif version.startswith("NRF54L10"):
            return "nRF54L10"
        elif version.startswith("NRF54L05"):
            return "nRF54L05"
        else:
            raise NotImplementedError(f"Unhandled device {version}")


class nRF91(NRFFamily):
    FICR_ADDRESS = 0x00FF0000
    DEVICE_ID_OFFSET = 0x204
    CUSTOMER_OFFSET = 0x108

    @staticmethod
    def soc(device_info):
        version: str = device_info["jlink"]["deviceVersion"]
        if version.startswith("NRF9160"):
            return "nRF9160"
        elif version.startswith("NRF9161"):
            return "nRF9161"
        elif version.startswith("NRF9151"):
            return "nRF9151"
        else:
            raise NotImplementedError(f"Unhandled device {version}")


DEVICE_FAMILY_MAPPING: dict[str, type[NRFFamily]] = {
    "NRF52_FAMILY": nRF52,
    "NRF53_FAMILY": nRF53,
    "NRF54L_FAMILY": nRF54L,
    "NRF91_FAMILY": nRF91,
}


class Interface(ProvisioningInterface):
    def __init__(self, snr: int | None):
        self.snr = snr
        devices = self._exec(["device-info"])
        if len(devices) == 0:
            raise RuntimeError("No devices found")
        devices_info = devices[0]["devices"]

        if snr is None:
            if len(devices_info) > 1:
                serials = ",".join([d["serialNumber"] for d in devices_info])
                raise RuntimeError(f"Multiple devices found without a SNR provided (Found: {serials})")
            self.snr = int(devices_info[0]["serialNumber"])
        infos = [info for info in devices_info if int(info["serialNumber"]) == self.snr]
        if len(infos) == 0:
            raise RuntimeError(f"Devices with SNR {self.snr} not found")
        self.device_info = infos[0]["deviceInfo"]
        self.core_info = self._exec(["core-info"])
        self.family = DEVICE_FAMILY_MAPPING[self.device_info["jlink"]["deviceFamily"]]
        self.uicr_base: int = self.core_info[0]["devices"][0]["uicrAddress"]
        self._soc_name: str = self.family.soc(self.device_info)

    def _exec(self, args: list[str]):
        jout_all = []
        cmd_base = ["nrfutil", "--json", "--skip-overhead", "device"]
        cmd = cmd_base + args
        if self.snr is not None:
            cmd += ["--serial-number", str(self.snr)]

        with subprocess.Popen(cmd, stdout=subprocess.PIPE) as p:
            assert p.stdout is not None
            for line in iter(p.stdout.readline, b""):
                # https://github.com/ndjson/ndjson-spec
                jout = json.loads(line.decode(sys.getdefaultencoding()))
                jout_all.append(jout)

        return jout_all

    def close(self):
        self._exec(["reset"])

    @property
    def soc_name(self) -> str:
        return self._soc_name

    @property
    def unique_device_id_len(self) -> int:
        return 8

    def unique_device_id(self) -> int:
        device_id_addr = self.family.FICR_ADDRESS + self.family.DEVICE_ID_OFFSET

        result = self._exec(["read", "--address", hex(device_id_addr), "--bytes", "8", "--direct"])
        data_bytes = result[0]["devices"][0]["memoryData"][0]["values"]
        dev_id_bytes = bytes(data_bytes)
        return int.from_bytes(dev_id_bytes, "big")

    def read_provisioned_data(self, num: int) -> bytes:
        customer_addr = self.uicr_base + self.family.CUSTOMER_OFFSET

        result = self._exec(["read", "--address", hex(customer_addr), "--bytes", str(num), "--direct"])
        data_bytes = result[0]["devices"][0]["memoryData"][0]["values"]

        return bytes(data_bytes)

    def write_provisioning_data(self, data: bytes):
        customer_addr = self.uicr_base + self.family.CUSTOMER_OFFSET

        # x-write only operates on single words
        for offset in range(0, len(data), 4):
            chunk_bytes = data[offset : offset + 4]
            if len(chunk_bytes) != 4:
                chunk_bytes += b"\xff" * (4 - len(chunk_bytes))
            data_word = int.from_bytes(chunk_bytes, byteorder="little")

            self._exec(["write", "--address", hex(customer_addr + offset), "--value", hex(data_word)])



from http import HTTPStatus
from typing import Any

import httpx

from ... import errors
from ...client import AuthenticatedClient, Client
from ...models.derive_device_key_body import DeriveDeviceKeyBody
from ...models.error import Error
from ...models.key import Key
from ...types import Response


def _get_kwargs(
    *,
    body: DeriveDeviceKeyBody,
) -> dict[str, Any]:
    headers: dict[str, Any] = {}

    _kwargs: dict[str, Any] = {
        "method": "post",
        "url": "/key/derived/device",
    }

    _body = body.to_dict()

    _kwargs["json"] = _body
    headers["Content-Type"] = "application/json"

    _kwargs["headers"] = headers
    return _kwargs


def _parse_response(*, client: AuthenticatedClient | Client, response: httpx.Response) -> Error | Key | None:
    if response.status_code == 200:
        response_200 = Key.from_dict(response.json())

        return response_200
    if response.status_code == 400:
        response_400 = Error.from_dict(response.json())

        return response_400
    if client.raise_on_unexpected_status:
        raise errors.UnexpectedStatus(response.status_code, response.content)
    else:
        return None


def _build_response(*, client: AuthenticatedClient | Client, response: httpx.Response) -> Response[Error | Key]:
    return Response(
        status_code=HTTPStatus(response.status_code),
        content=response.content,
        headers=response.headers,
        parsed=_parse_response(client=client, response=response),
    )


def sync_detailed(
    *,
    client: AuthenticatedClient | Client,
    body: DeriveDeviceKeyBody,
) -> Response[Error | Key]:
    """Derive a device key for encryption

     Generate a derived key to use for device level encrpytion, if security state is provided, it will be
    used to derive the key, otherwise the last stored security state will be used.

    Args:
        body (DeriveDeviceKeyBody):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[Error, Key]]
    """

    kwargs = _get_kwargs(
        body=body,
    )

    response = client.get_httpx_client().request(
        **kwargs,
    )

    return _build_response(client=client, response=response)


def sync(
    *,
    client: AuthenticatedClient | Client,
    body: DeriveDeviceKeyBody,
) -> Error | Key | None:
    """Derive a device key for encryption

     Generate a derived key to use for device level encrpytion, if security state is provided, it will be
    used to derive the key, otherwise the last stored security state will be used.

    Args:
        body (DeriveDeviceKeyBody):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[Error, Key]
    """

    return sync_detailed(
        client=client,
        body=body,
    ).parsed


async def asyncio_detailed(
    *,
    client: AuthenticatedClient | Client,
    body: DeriveDeviceKeyBody,
) -> Response[Error | Key]:
    """Derive a device key for encryption

     Generate a derived key to use for device level encrpytion, if security state is provided, it will be
    used to derive the key, otherwise the last stored security state will be used.

    Args:
        body (DeriveDeviceKeyBody):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[Error, Key]]
    """

    kwargs = _get_kwargs(
        body=body,
    )

    response = await client.get_async_httpx_client().request(**kwargs)

    return _build_response(client=client, response=response)


async def asyncio(
    *,
    client: AuthenticatedClient | Client,
    body: DeriveDeviceKeyBody,
) -> Error | Key | None:
    """Derive a device key for encryption

     Generate a derived key to use for device level encrpytion, if security state is provided, it will be
    used to derive the key, otherwise the last stored security state will be used.

    Args:
        body (DeriveDeviceKeyBody):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[Error, Key]
    """

    return (
        await asyncio_detailed(
            client=client,
            body=body,
        )
    ).parsed



#!/usr/bin/env python3

from infuse_iot.generated.rpc_definitions import *  # noqa: F403
from infuse_iot.util.internal import extension_load as _loader

if _extension_module := _loader("rpc_definitions"):
    _globals = globals()
    for _name in _extension_module.__all__:
        _globals[_name] = getattr(_extension_module, _name)



#!/usr/bin/env python3

"""Infuse-IoT SDK meta-tool command parent class"""

__author__ = "Jordan Yates"
__copyright__ = "Copyright 2024, Embeint Holdings Pty Ltd"

import argparse
import ctypes
from abc import ABCMeta, abstractmethod
from typing import Any

from infuse_iot.epacket.packet import Auth


class InfuseCommand(metaclass=ABCMeta):
    """Infuse-IoT SDK meta-tool command parent class"""

    @classmethod
    def add_parser(cls, parser: argparse.ArgumentParser):
        """Add arguments for sub-command"""
        return

    def __init__(self, args: argparse.Namespace):
        return

    @abstractmethod
    def run(self) -> None:
        """Run the subcommand"""

    @property
    @abstractmethod
    def NAME(self) -> str:
        pass

    @property
    @abstractmethod
    def HELP(self) -> str:
        pass

    @property
    @abstractmethod
    def DESCRIPTION(self) -> str:
        pass


class InfuseRpcCommand:
    RPC_DATA_SEND: bool = False
    RPC_DATA_SEND_CHUNKED: bool = False
    RPC_DATA_RECEIVE: bool = False

    @classmethod
    def add_parser(cls, parser: argparse.ArgumentParser) -> None:
        raise NotImplementedError

    def __init__(self, **kwargs):
        pass

    def auth_level(self) -> Auth:
        """Authentication level to run command with"""
        return Auth.DEVICE

    def request_struct(self) -> ctypes.LittleEndianStructure:
        """RPC_CMD request structure"""
        raise NotImplementedError

    def request_json(self) -> dict[str, Any]:
        """RPC_CMD json structure (cloud)"""
        raise NotImplementedError

    def data_payload(self) -> bytes:
        """Payload to send with RPC_DATA"""
        raise NotImplementedError

    def data_payload_chunked(self) -> list[bytes]:
        """Payloads to send with RPC_DATA"""
        raise NotImplementedError

    def data_payload_recv_len(self) -> int:
        """Length of payload to receive with RPC_DATA"""
        return 0xFFFFFFFF

    def data_recv_cb(self, offset: int, data: bytes) -> None:
        """Data received callback"""
        raise NotImplementedError

    def data_progress_cb(self, offset: int) -> None:
        """Progress callback"""
        raise NotImplementedError

    def handle_response(self, return_code: int, response: ctypes.LittleEndianStructure | None) -> None:
        """Handle RPC_RSP"""
        raise NotImplementedError



#!/usr/bin/env python3

import infuse_iot.definitions.rpc as defs
from infuse_iot.commands import InfuseRpcCommand
from infuse_iot.zephyr.errno import errno


class gravity_reference_update(InfuseRpcCommand, defs.gravity_reference_update):
    @classmethod
    def add_parser(cls, parser):
        parser.add_argument("--variance", type=int, default=0, help="Maximum axis variance to accept")

    def __init__(self, args):
        self._max_variance = args.variance

    def request_struct(self):
        return self.request(self._max_variance)

    def request_json(self):
        return {"max_variance": str(self._max_variance)}

    def handle_response(self, return_code, response):
        r = response

        if return_code == -errno.EIO:
            print(f"IMU variance too large: {r.variance.x:6d} {r.variance.y:6d} {r.variance.z:6d}")
            return
        elif return_code < 0:
            print(f"Failed to update gravity reference vector ({errno.strerror(-return_code)})")
            return

        print(f"\t  Gravity: {r.reference.x:6d} {r.reference.y:6d} {r.reference.z:6d}")
        print(f"\t Variance: {r.variance.x:6d} {r.variance.y:6d} {r.variance.z:6d}")
        print(f"\t  Samples: {r.num_samples:d}")
        print(f"\t   Period: {r.sample_period_us:d} us")



#!/usr/bin/env python3

import ctypes

import infuse_iot.definitions.rpc as defs
from infuse_iot.commands import InfuseRpcCommand
from infuse_iot.util.ctypes import VLACompatLittleEndianStruct
from infuse_iot.zephyr.errno import errno


class lte_at_cmd(InfuseRpcCommand, defs.lte_at_cmd):
    class request(ctypes.LittleEndianStructure):
        _pack_ = 1

    class response(VLACompatLittleEndianStruct):
        vla_field = ("rsp", 0 * ctypes.c_char)

    @classmethod
    def add_parser(cls, parser):
        parser.add_argument("--cmd", "-c", type=str, help="Command string")

    def __init__(self, args):
        self.args = args

    def request_struct(self):
        return self.args.cmd.encode("utf-8") + b"\x00"

    def request_json(self):
        return {"cmd": self.args.cmd}

    def handle_response(self, return_code, response):
        if response:
            # Print returned strings even on failure
            response_bytes = bytes(response.rsp)
            if len(response_bytes):
                decoded = bytes(response.rsp).decode("utf-8").strip()
                print(decoded)
        # Notification that command failed
        if return_code != 0:
            print(f"Failed to run command ({errno.strerror(-return_code)})")



#!/usr/bin/env python3


import infuse_iot.definitions.rpc as defs
from infuse_iot.commands import InfuseRpcCommand
from infuse_iot.zephyr.errno import errno


class time_get(InfuseRpcCommand, defs.time_get):
    @classmethod
    def add_parser(cls, parser):
        pass

    def __init__(self, args):
        pass

    def request_struct(self):
        return self.request()

    def request_json(self):
        return {}

    def handle_response(self, return_code, response):
        if return_code != 0:
            print(f"Failed to query current time ({errno.strerror(-return_code)})")
            return

        import time

        from infuse_iot.time import InfuseTime, InfuseTimeSource

        t_remote = InfuseTime.unix_time_from_epoch(response.epoch_time)
        t_local = time.time()
        sync_age = f"{response.sync_age} seconds ago" if response.sync_age != 2**32 - 1 else "Never"

        print(f"\t     Source: {InfuseTimeSource(response.time_source)}")
        print(f"\tRemote Time: {InfuseTime.utc_time_string(t_remote)}")
        print(f"\t Local Time: {InfuseTime.utc_time_string(t_local)}")
        print(f"\t     Synced: {sync_age}")



#!/usr/bin/env python3

"""Connect to remote Bluetooth device serial logs"""

__author__ = "Jordan Yates"
__copyright__ = "Copyright 2024, Embeint Holdings Pty Ltd"

import time
import wave
from contextlib import ExitStack

from infuse_iot.commands import InfuseCommand
from infuse_iot.common import InfuseID, InfuseType
from infuse_iot.definitions import tdf as tdf_defs
from infuse_iot.epacket import interface
from infuse_iot.socket_comms import (
    ClientNotificationConnectionDropped,
    ClientNotificationEpacketReceived,
    GatewayRequestConnectionRequest,
    LocalClient,
    default_multicast_address,
)
from infuse_iot.tdf import TDF
from infuse_iot.util.console import Console


class SubCommand(InfuseCommand):
    NAME = "audio_record"
    HELP = "Record audio data to a file from TDF"
    DESCRIPTION = "Record audio data to a file from TDF"

    def __init__(self, args):
        self._client = LocalClient(default_multicast_address(), 1.0)
        self._decoder = TDF()
        if args.gateway:
            self._id = InfuseID.GATEWAY
        else:
            self._id = args.id
        self._file_prefix = f"{args.name}_" if args.name else ""
        self._conn_timeout = args.conn_timeout
        self._freq: int | None = None
        self._left: wave.Wave_write | None = None
        self._right: wave.Wave_write | None = None

    @classmethod
    def add_parser(cls, parser):
        addr_group = parser.add_mutually_exclusive_group(required=True)
        addr_group.add_argument("--gateway", action="store_true", help="Run command on local gateway")
        addr_group.add_argument("--id", type=lambda x: int(x, 0), help="Infuse ID to run command on")
        parser.add_argument(
            "--conn-timeout", type=int, default=10000, help="Timeout to wait for a connection to the device (ms)"
        )
        parser.add_argument("--name", type=str, help="Filename prefix")

    def handle_channel(self, channel: str, stack: ExitStack, tdf: TDF.Reading):
        if channel == "left":
            chan = self._left
        else:
            chan = self._right

        if chan is None:
            filename = f"{self._file_prefix}{int(time.time())}_{channel}.wav"
            Console.log_info(f"Opening '{filename}'")
            chan = stack.enter_context(wave.open(filename, "wb"))  # noqa: SIM115
            chan.setnchannels(1)
            chan.setsampwidth(2)
            assert self._freq
            chan.setframerate(float(self._freq))

            if channel == "left":
                self._left = chan
            else:
                self._right = chan

        samples = b"".join([x.val.to_bytes(2, "little", signed=True) for x in tdf.data])
        chan.writeframes(samples)

    def handle_connection(self):
        with ExitStack() as stack:
            Console.log_info("Waiting for frequency information...")
            while evt := self._client.receive():
                if evt is None:
                    continue
                if isinstance(evt, ClientNotificationConnectionDropped):
                    Console.log_error(f"Connection to {self._id:016x} lost")
                    break
                if not isinstance(evt, ClientNotificationEpacketReceived):
                    continue
                source = evt.epacket.route[0]
                if source.infuse_id != self._id:
                    continue
                if source.interface != interface.ID.BT_CENTRAL:
                    continue
                if evt.epacket.ptype != InfuseType.TDF:
                    continue
                for tdf in self._decoder.decode(evt.epacket.payload):
                    if self._freq is None:
                        if tdf.id == tdf_defs.readings.idx_array_freq.ID:
                            self._freq = tdf.data[0].frequency
                            Console.log_info(f"Audio frequency is {self._freq} Hz")
                        else:
                            # Don't write until metadata is known
                            continue
                    if tdf.id == tdf_defs.readings.pcm_16bit_chan_left.ID:
                        self.handle_channel("left", stack, tdf)
                    elif tdf.id == tdf_defs.readings.pcm_16bit_chan_right.ID:
                        self.handle_channel("right", stack, tdf)

    def run(self):
        try:
            types = GatewayRequestConnectionRequest.DataType.DATA
            Console.log_info(f"Connecting to 0x{self._id:016x}")
            with self._client.connection(self._id, types, self._conn_timeout) as _:
                self.handle_connection()

        except KeyboardInterrupt:
            Console.log_error(f"Disconnecting from {self._id:016x}")
        except ConnectionRefusedError:
            Console.log_error(f"Unable to connect to {self._id:016x}")

        if self._left:
            assert self._freq
            Console.log_text(f"Left Channel Recorded: {self._left.getnframes() / self._freq} Seconds")
        if self._right:
            assert self._freq
            Console.log_text(f"Right Channel Recorded: {self._right.getnframes() / self._freq} Seconds")



#!/usr/bin/env python3

import ctypes
import ipaddress
import socket

import tabulate

import infuse_iot.definitions.rpc as defs
from infuse_iot.commands import InfuseRpcCommand
from infuse_iot.zephyr.errno import errno
from infuse_iot.zephyr.net import AddressFamily, SockType


class zperf_upload(InfuseRpcCommand, defs.zperf_upload):
    @classmethod
    def add_parser(cls, parser):
        parser.add_argument("--address", "-a", type=str, required=True, help="Peer IP address")
        parser.add_argument("--port", "-p", type=int, default=5001, help="Peer port")
        socket_group = parser.add_mutually_exclusive_group(required=True)
        socket_group.add_argument(
            "--tcp",
            dest="sock_type",
            action="store_const",
            const=SockType.SOCK_STREAM,
            help="TCP protocol",
        )
        socket_group.add_argument(
            "--udp",
            dest="sock_type",
            action="store_const",
            const=SockType.SOCK_DGRAM,
            help="UDP protocol",
        )
        source_group = parser.add_mutually_exclusive_group()
        source_group.add_argument(
            "--constant",
            dest="data_source",
            action="store_const",
            const=defs.rpc_enum_zperf_data_source.CONSTANT,
            default=defs.rpc_enum_zperf_data_source.CONSTANT,
            help="Constant data payload ('z')",
        )
        source_group.add_argument(
            "--random",
            dest="data_source",
            action="store_const",
            const=defs.rpc_enum_zperf_data_source.RANDOM,
            help="Random data payload",
        )
        source_group.add_argument(
            "--onboard",
            dest="data_source",
            action="store_const",
            const=defs.rpc_enum_zperf_data_source.FLASH_ONBOARD,
            help="Read from onboard flash logger",
        )
        source_group.add_argument(
            "--removable",
            dest="data_source",
            action="store_const",
            const=defs.rpc_enum_zperf_data_source.FLASH_REMOVABLE,
            help="Read from removable flash logger",
        )
        parser.add_argument("--encrypt", action="store_true", help="Encrypt payloads before transmission")
        parser.add_argument("--duration", type=int, default=5000, help="Duration to run test over in milliseconds")
        parser.add_argument("--rate-kbps", type=int, default=0, help="Desired upload rate in kbps")
        parser.add_argument("--payload-size", type=int, default=512, help="Payload size")

    def __init__(self, args):
        self.peer_addr = ipaddress.ip_address(args.address)
        self.peer_port = args.port
        self.sock_type = args.sock_type
        self.data_source = args.data_source
        if args.encrypt:
            self.data_source |= defs.rpc_enum_zperf_data_source.ENCRYPT
        self.duration = args.duration
        self.rate = args.rate_kbps
        self.packet_size = args.payload_size
        if self.sock_type == SockType.SOCK_DGRAM:
            # Add the UDP client header size to the requested payload size
            self.packet_size += 40

    def request_struct(self):
        peer_family = (
            AddressFamily.AF_INET if isinstance(self.peer_addr, ipaddress.IPv4Address) else AddressFamily.AF_INET6
        )
        addr_bytes = (16 * ctypes.c_uint8)(*self.peer_addr.packed)
        peer = defs.rpc_struct_sockaddr(
            sin_family=peer_family,
            sin_port=socket.htons(self.peer_port),
            sin_addr=addr_bytes,
            scope_id=0,
        )

        return self.request(
            peer_address=peer,
            sock_type=self.sock_type,
            data_source=self.data_source,
            duration_ms=self.duration,
            rate_kbps=self.rate,
            packet_size=self.packet_size,
        )

    def request_json(self):
        return {}

    def handle_response(self, return_code, response):
        if return_code != 0:
            print(f"Failed to run zperf ({errno.strerror(-return_code)})")
            return

        throughput_bps = 8 * response.total_len / (response.client_time_in_us / 1000)
        print(f"Average Throughput: {throughput_bps / 1000:.3f} kbps")
        if self.sock_type == SockType.SOCK_DGRAM:
            recv = 100 * response.nb_packets_rcvd / response.nb_packets_sent
            loss = 100 * response.nb_packets_lost / response.nb_packets_sent
            print(f"       Packet Recv: {recv:6.2f}%")
            print(f"       Packet Loss: {loss:6.2f}%")
        results = []
        for field_name, _ in response._fields_:
            results.append([field_name, getattr(response, field_name)])
        print(tabulate.tabulate(results))



from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, TypeVar
from uuid import UUID

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.metadata_field import MetadataField


T = TypeVar("T", bound="NewBoard")


@_attrs_define
class NewBoard:
    """
    Attributes:
        name (str): Name of board Example: Board Name.
        description (str): Description of board Example: Extended description of board.
        soc (str): System on Chip (SoC) of board Example: nRF9151.
        organisation_id (UUID): ID of organisation for board to exist in
        metadata_fields (Union[Unset, list['MetadataField']]): Metadata fields for board Example: [{'name': 'Field
            Name', 'required': True, 'unique': False}].
    """

    name: str
    description: str
    soc: str
    organisation_id: UUID
    metadata_fields: Unset | list["MetadataField"] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        name = self.name

        description = self.description

        soc = self.soc

        organisation_id = str(self.organisation_id)

        metadata_fields: Unset | list[dict[str, Any]] = UNSET
        if not isinstance(self.metadata_fields, Unset):
            metadata_fields = []
            for componentsschemas_board_metadata_fields_item_data in self.metadata_fields:
                componentsschemas_board_metadata_fields_item = (
                    componentsschemas_board_metadata_fields_item_data.to_dict()
                )
                metadata_fields.append(componentsschemas_board_metadata_fields_item)

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "name": name,
                "description": description,
                "soc": soc,
                "organisationId": organisation_id,
            }
        )
        if metadata_fields is not UNSET:
            field_dict["metadataFields"] = metadata_fields

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.metadata_field import MetadataField

        d = dict(src_dict)
        name = d.pop("name")

        description = d.pop("description")

        soc = d.pop("soc")

        organisation_id = UUID(d.pop("organisationId"))

        metadata_fields = []
        _metadata_fields = d.pop("metadataFields", UNSET)
        for componentsschemas_board_metadata_fields_item_data in _metadata_fields or []:
            componentsschemas_board_metadata_fields_item = MetadataField.from_dict(
                componentsschemas_board_metadata_fields_item_data
            )

            metadata_fields.append(componentsschemas_board_metadata_fields_item)

        new_board = cls(
            name=name,
            description=description,
            soc=soc,
            organisation_id=organisation_id,
            metadata_fields=metadata_fields,
        )

        new_board.additional_properties = d
        return new_board

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



#!/usr/bin/env python3


import infuse_iot.definitions.rpc as defs
from infuse_iot.commands import InfuseRpcCommand
from infuse_iot.util.time import humanised_seconds
from infuse_iot.zephyr.errno import errno


class data_logger_state(InfuseRpcCommand, defs.data_logger_state):
    @classmethod
    def add_parser(cls, parser):
        logger = parser.add_mutually_exclusive_group(required=True)
        logger.add_argument("--onboard", action="store_true", help="Onboard flash logger")
        logger.add_argument("--removable", action="store_true", help="Removable flash logger (SD)")

    def __init__(self, args):
        if args.onboard:
            self.logger = defs.rpc_enum_data_logger.FLASH_ONBOARD
        elif args.removable:
            self.logger = defs.rpc_enum_data_logger.FLASH_REMOVABLE
        else:
            raise NotImplementedError

    def request_struct(self):
        return self.request(self.logger)

    def request_json(self):
        return {"logger": self.logger.name}

    def handle_response(self, return_code, response):
        if return_code != 0:
            print(f"Failed to query data logger state ({errno.strerror(-return_code)})")
            return

        def sizeof_fmt(num, suffix="B"):
            for unit in ("", "Ki", "Mi", "Gi", "Ti", "Pi", "Ei", "Zi"):
                if abs(num) < 1024.0:
                    return f"{num:3.2f} {unit}{suffix}"
                num /= 1024.0
            return f"{num:.2f} Yi{suffix}"

        r = response
        total_logged = r.current_block * r.block_size
        percent = 100 * r.current_block / r.logical_blocks
        block_rate = (r.current_block - r.boot_block) / r.uptime
        byte_rate = r.bytes_logged / r.uptime

        print(f"{self.logger.name}")
        print(f"\t     Logged: {sizeof_fmt(total_logged)}")
        print(f"\t     Blocks: {r.current_block}/{r.logical_blocks} ({percent:.0f}%)")
        if r.earliest_block > 0:
            print(f"\t   Earliest: {r.earliest_block}")
        if byte_rate == 0.0:
            print("\t Block Rate: N/A")
            print("\t  Byte Rate: N/A")
        elif byte_rate < 0.1:
            print(f"\t Block Rate: {1 / block_rate:.2f} sec/block")
            print(f"\t  Byte Rate: {1 / byte_rate:.2f} sec/byte")
        else:
            print(f"\t Block Rate: {block_rate:.2f} blocks/sec")
            print(f"\t  Byte Rate: {sizeof_fmt(byte_rate)}/sec")
        if r.bytes_logged > 0:
            physical_wrap_time = r.physical_blocks / block_rate
            logical_fill_time = r.logical_blocks / block_rate
            print(f"\t   Phy Wrap: {humanised_seconds(int(physical_wrap_time))}")
            print(f"\t   Log Fill: {humanised_seconds(int(logical_fill_time))}")



from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, TypeVar

from attrs import define as _attrs_define
from attrs import field as _attrs_field

if TYPE_CHECKING:
    from ..models.uplink_route import UplinkRoute


T = TypeVar("T", bound="UplinkRouteAndDeviceId")


@_attrs_define
class UplinkRouteAndDeviceId:
    """Uplink route with device ID

    Attributes:
        device_id (str): 8 byte DeviceID as a hex string Example: d291d4d66bf0a955.
        last_route (UplinkRoute):
    """

    device_id: str
    last_route: "UplinkRoute"
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        device_id = self.device_id

        last_route = self.last_route.to_dict()

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "deviceId": device_id,
                "lastRoute": last_route,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.uplink_route import UplinkRoute

        d = dict(src_dict)
        device_id = d.pop("deviceId")

        last_route = UplinkRoute.from_dict(d.pop("lastRoute"))

        uplink_route_and_device_id = cls(
            device_id=device_id,
            last_route=last_route,
        )

        uplink_route_and_device_id.additional_properties = d
        return uplink_route_and_device_id

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..models.key_interface import KeyInterface
from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.security_state import SecurityState


T = TypeVar("T", bound="DeriveDeviceKeyBody")


@_attrs_define
class DeriveDeviceKeyBody:
    """
    Attributes:
        device_id (str): The ID of the device to send the RPC to as a hex string Example: d291d4d66bf0a955.
        interface (KeyInterface):
        security_state (Union[Unset, SecurityState]):
    """

    device_id: str
    interface: KeyInterface
    security_state: Union[Unset, "SecurityState"] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        device_id = self.device_id

        interface = self.interface.value

        security_state: Unset | dict[str, Any] = UNSET
        if not isinstance(self.security_state, Unset):
            security_state = self.security_state.to_dict()

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "deviceId": device_id,
                "interface": interface,
            }
        )
        if security_state is not UNSET:
            field_dict["securityState"] = security_state

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.security_state import SecurityState

        d = dict(src_dict)
        device_id = d.pop("deviceId")

        interface = KeyInterface(d.pop("interface"))

        _security_state = d.pop("securityState", UNSET)
        security_state: Unset | SecurityState
        if isinstance(_security_state, Unset):
            security_state = UNSET
        else:
            security_state = SecurityState.from_dict(_security_state)

        derive_device_key_body = cls(
            device_id=device_id,
            interface=interface,
            security_state=security_state,
        )

        derive_device_key_body.additional_properties = d
        return derive_device_key_body

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



#!/usr/bin/env python3

"""Test serial throughput to local gateway"""

__author__ = "Jordan Yates"
__copyright__ = "Copyright 2024, Embeint Holdings Pty Ltd"

import random
import time

from infuse_iot.commands import InfuseCommand
from infuse_iot.common import InfuseID, InfuseType
from infuse_iot.epacket.packet import Auth, PacketOutput
from infuse_iot.socket_comms import (
    ClientNotificationEpacketReceived,
    GatewayRequestEpacketSend,
    LocalClient,
    default_multicast_address,
)


class SubCommand(InfuseCommand):
    NAME = "serial_throughput"
    HELP = "Test serial throughput to local gateway"
    DESCRIPTION = "Test serial throughput to local gateway"

    @classmethod
    def add_parser(cls, parser):
        parser.add_argument(
            "--iterations",
            type=int,
            default=20,
            help="Number of times to send each sized packet",
        )

    def __init__(self, args):
        self._client = LocalClient(default_multicast_address(), 1.0)
        self._iterations = args.iterations

    def run_send_test(self, num, size, queue_size):
        assert size >= 4
        self._client.set_rx_timeout(0.2)
        sent = 0
        pending = 0
        responses = 0

        start = time.time()
        while (sent != num) or (pending > 0):
            # Queue packets up to the maximum queue size
            while (sent != num) and (pending < queue_size):
                payload = sent.to_bytes(4, "little") + random.randbytes(size - 4)
                pkt = PacketOutput(
                    InfuseID.GATEWAY,
                    Auth.DEVICE,
                    InfuseType.ECHO_REQ,
                    payload,
                )
                req = GatewayRequestEpacketSend(pkt)
                self._client.send(req)
                sent += 1
                pending += 1
            # Wait for responses
            if rsp := self._client.receive():
                if not isinstance(rsp, ClientNotificationEpacketReceived):
                    continue
                if rsp.epacket.ptype != InfuseType.ECHO_RSP:
                    continue
                responses += 1
                pending -= 1
            else:
                # Timeout, no more data coming
                break
        end = time.time()
        duration = end - start
        total = num * size
        throughput = total / duration
        if responses != num:
            print(f"\tOnly received {responses}/{num} responses")
        msg = f"\t{num} packets with {size:3d} bytes payload complete in {duration:.2f} seconds"
        msg += f" ({int(8 * throughput):6d} bps)"
        print(msg)

    def run(self):
        # No queuing
        print(f"Averaged across {self._iterations} packets with no queuing:")
        self.run_send_test(self._iterations, 4, 1)
        self.run_send_test(self._iterations, 32, 1)
        self.run_send_test(self._iterations, 128, 1)
        self.run_send_test(self._iterations, 512, 1)
        # Small queue
        print(f"Averaged across {self._iterations} packets with 3 packet queue:")
        self.run_send_test(self._iterations, 4, 3)
        self.run_send_test(self._iterations, 32, 3)
        self.run_send_test(self._iterations, 128, 3)
        self.run_send_test(self._iterations, 512, 3)
        # Larger queue
        print(f"Averaged across {self._iterations} packets with 5 packet queue:")
        self.run_send_test(self._iterations, 4, 5)
        self.run_send_test(self._iterations, 32, 5)
        self.run_send_test(self._iterations, 128, 5)
        self.run_send_test(self._iterations, 512, 5)



#!/usr/bin/env python

import importlib.util
import os
import pathlib
import types

import infuse_iot.credentials


def extension_load(name: str) -> None | types.ModuleType:
    if os.environ.get("_ARGCOMPLETE") == "1":
        # Skip expensive imports when running argcomplete
        return None

    defs_path = infuse_iot.credentials.get_custom_definitions_path()
    if defs_path is None:
        return None
    extensions_file = pathlib.Path(defs_path) / f"{name}.py"
    if not extensions_file.exists():
        return None

    try:
        # Import the extension file
        spec = importlib.util.spec_from_file_location(f"infuse_iot.extension.{name}", str(extensions_file))
        if spec is None or spec.loader is None:
            return None
        file_module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(file_module)
    except Exception as _:
        return None

    # Return the loaded file
    return file_module



#!/usr/bin/env python3


import infuse_iot.definitions.rpc as defs
from infuse_iot.commands import InfuseRpcCommand
from infuse_iot.zephyr.errno import errno


class time_set(InfuseRpcCommand, defs.time_set):
    @classmethod
    def add_parser(cls, parser):
        pass

    def __init__(self, args):
        pass

    def request_struct(self):
        import time

        from infuse_iot.time import InfuseTime

        return self.request(InfuseTime.epoch_time_from_unix(time.time()))

    def handle_response(self, return_code, response):
        if return_code != 0:
            print(f"Failed to set current time ({errno.strerror(-return_code)})")
            return
        else:
            print("Set current time on device")



import datetime
from collections.abc import Mapping
from typing import Any, TypeVar
from uuid import UUID

from attrs import define as _attrs_define
from attrs import field as _attrs_field
from dateutil.parser import isoparse

T = TypeVar("T", bound="CreatedDeviceProperties")


@_attrs_define
class CreatedDeviceProperties:
    """
    Attributes:
        id (UUID): Generated UUID for organisation
        created_at (datetime.datetime):
        updated_at (datetime.datetime):
    """

    id: UUID
    created_at: datetime.datetime
    updated_at: datetime.datetime
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        id = str(self.id)

        created_at = self.created_at.isoformat()

        updated_at = self.updated_at.isoformat()

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "id": id,
                "createdAt": created_at,
                "updatedAt": updated_at,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        d = dict(src_dict)
        id = UUID(d.pop("id"))

        created_at = isoparse(d.pop("createdAt"))

        updated_at = isoparse(d.pop("updatedAt"))

        created_device_properties = cls(
            id=id,
            created_at=created_at,
            updated_at=updated_at,
        )

        created_device_properties.additional_properties = d
        return created_device_properties

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



#!/usr/bin/env python3

"""Provision device on Infuse Cloud"""

__author__ = "Jordan Yates"
__copyright__ = "Copyright 2024, Embeint Holdings Pty Ltd"

import ctypes
import sys
from http import HTTPStatus

from infuse_iot.api_client import Client
from infuse_iot.api_client.api.board import get_board_by_id, get_boards
from infuse_iot.api_client.api.device import (
    create_device,
    get_device_by_soc_and_mcu_id,
)
from infuse_iot.api_client.api.organisation import get_all_organisations
from infuse_iot.api_client.models import Board, DeviceMetadata, Error, NewDevice
from infuse_iot.commands import InfuseCommand
from infuse_iot.credentials import get_api_key
from infuse_iot.util.console import choose_one
from infuse_iot.util.soc import nrf, soc, stm


class SubCommand(InfuseCommand):
    NAME = "provision"
    HELP = "Provision device on Infuse Cloud"
    DESCRIPTION = "Provision device on Infuse Cloud"

    @classmethod
    def add_parser(cls, parser):
        vendor_group = parser.add_mutually_exclusive_group(required=True)
        vendor_group.add_argument(
            "--nrf", dest="vendor", action="store_const", const="nrf", help="Nordic Semiconductor SoC"
        )
        vendor_group.add_argument(
            "--stm", dest="vendor", action="store_const", const="stm", help="ST Microelectronics SoC"
        )
        parser.add_argument(
            "--snr",
            type=int,
            default=None,
            help="JTAG serial number",
        )
        parser.add_argument("--board", "-b", type=str, help="Board ID")
        parser.add_argument("--organisation", "-o", type=str, help="Organisation ID")
        parser.add_argument(
            "--id",
            "-i",
            type=lambda x: int(x, 0),
            help="Infuse device ID to provision as",
        )
        parser.add_argument(
            "--metadata",
            "-m",
            metavar="KEY=VALUE",
            nargs="+",
            type=str,
            help="Define a number of key-value pairs for metadata",
        )
        parser.add_argument(
            "--dry-run", action="store_true", help="Generate the request that would be sent, but do not send it"
        )

    def __init__(self, args):
        self._vendor = args.vendor
        self._snr = args.snr
        self._board = args.board
        self._org = args.organisation
        self._id = args.id
        self._dry_run = args.dry_run
        self._metadata = {}
        if args.metadata:
            for meta in args.metadata:
                key, val = meta.strip().split("=", 1)
                self._metadata[key.strip()] = val

    def create_device(self, client: Client, soc_name: str, hardware_id_str: str):
        if self._org is None:
            orgs = get_all_organisations.sync(client=client)
            if isinstance(orgs, Error) or orgs is None:
                sys.exit(f"Organisation query failed {orgs}")
            options = [f"{o.name:20s} ({o.id})" for o in orgs]

            idx, _val = choose_one("Organisation", options)
            self._org = orgs[idx].id

        if self._board is None:
            boards = get_boards.sync(client=client, organisation_id=self._org)
            if isinstance(boards, Error) or boards is None:
                sys.exit(f"Board query failed {boards}")
            options = [f"{b.name:20s} ({b.id})" for b in boards]

            idx, _val = choose_one("Board", options)
            self._board = boards[idx].id
        board = get_board_by_id.sync(client=client, id=self._board)
        if not isinstance(board, Board):
            sys.exit(f"Board query failed {board}")
        if board.soc != soc_name:
            sys.exit(f"Found SoC '{soc_name}' but board '{board.name}' has SoC '{board.soc}'")

        new_board = NewDevice(
            mcu_id=hardware_id_str,
            organisation_id=self._org,
            board_id=self._board,
            metadata=DeviceMetadata.from_dict(self._metadata),
        )
        if self._id:
            new_board.device_id = f"{self._id:016x}"

        if self._dry_run:
            print(new_board)
            return

        response = create_device.sync_detailed(client=client, body=new_board)
        if response.status_code != HTTPStatus.CREATED:
            sys.exit(f"Failed to create device:\n\t<{response.status_code}> {response.content.decode('utf-8')}")

    def run(self):
        interface: soc.ProvisioningInterface
        if self._vendor == "nrf":
            interface = nrf.Interface(self._snr)
        elif self._vendor == "stm":
            interface = stm.Interface()
        else:
            raise NotImplementedError(f"Unhandled vendor '{self._vendor}'")

        hardware_id = interface.unique_device_id()
        hardware_id_str = f"{hardware_id:0{2 * interface.unique_device_id_len}x}"

        client = Client(base_url="https://api.infuse-iot.com").with_headers({"x-api-key": f"Bearer {get_api_key()}"})

        # Get existing device or create new device
        with client as client:
            response = get_device_by_soc_and_mcu_id.sync_detailed(
                client=client, soc=interface.soc_name, mcu_id=hardware_id_str
            )
            if response.status_code == HTTPStatus.OK:
                # Device found, fall through
                pass
            elif response.status_code == HTTPStatus.NOT_FOUND:
                # Create new device here
                self.create_device(client, interface.soc_name, hardware_id_str)
                # Exit if dry run only
                if self._dry_run:
                    return
                # Query information back out
                response = get_device_by_soc_and_mcu_id.sync_detailed(
                    client=client, soc=interface.soc_name, mcu_id=hardware_id_str
                )
                if response.status_code != HTTPStatus.OK:
                    err = "Failed to query device after creation:\n"
                    err += f"\t<{response.status_code}> {response.content.decode('utf-8')}"
                    sys.exit(err)
                print("To provision more devices like this:")
                print(f"\t infuse provision --organisation {self._org} --board {self._board}")
            else:
                err = "Failed to query device information:\n"
                err += f"\t<{response.status_code}> {response.content.decode('utf-8')}"
                sys.exit(err)

        assert response.parsed is not None
        assert isinstance(response.parsed.device_id, str)
        # Compare current flash contents to desired flash contents
        cloud_id = int(response.parsed.device_id, 16)
        current_bytes = interface.read_provisioned_data(ctypes.sizeof(interface.DefaultProvisioningStruct))
        desired = interface.DefaultProvisioningStruct(cloud_id)
        desired_bytes = bytes(desired)

        if current_bytes == desired_bytes:
            print(f"HW ID 0x{hardware_id:016x} already provisioned as 0x{desired.device_id:016x}")
        else:
            if current_bytes != len(current_bytes) * b"\xff":
                print(f"HW ID 0x{hardware_id:016x} already has incorrect provisioning info, recover device")
                return

            interface.write_provisioning_data(bytes(desired))
            print(f"HW ID 0x{hardware_id:016x} now provisioned as 0x{desired.device_id:016x}")

        interface.close()



#!/usr/bin/env python3

import ctypes
import os
import sys

import infuse_iot.definitions.kv as kv_defs
import infuse_iot.definitions.rpc as defs
from infuse_iot.commands import InfuseRpcCommand
from infuse_iot.util.ctypes import VLACompatLittleEndianStruct, bytes_to_uint8
from infuse_iot.zephyr.errno import errno


class kv_write(InfuseRpcCommand, defs.kv_write):
    HELP = "Write an arbitrary kv value"
    DESCRIPTION = "Write an arbitrary kv value"

    class request(ctypes.LittleEndianStructure):
        _fields_ = [
            ("num", ctypes.c_uint8),
        ]
        _pack_ = 1

    class response(VLACompatLittleEndianStruct):
        vla_field = ("rc", 0 * ctypes.c_int16)

    @staticmethod
    def kv_store_value_factory(id, value_bytes):
        class kv_store_value(ctypes.LittleEndianStructure):
            _fields_ = [
                ("id", ctypes.c_uint16),
                ("len", ctypes.c_uint16),
                ("data", ctypes.c_ubyte * len(value_bytes)),
            ]
            _pack_ = 1

        return kv_store_value(id, len(value_bytes), bytes_to_uint8(value_bytes))

    @classmethod
    def add_parser(cls, parser):
        parser.add_argument("--key", "-k", type=int, required=True, help="KV key ID")
        v_parser = parser.add_mutually_exclusive_group(required=True)
        v_parser.add_argument("--value", "-v", type=str, help="KV value as hex string")
        v_parser.add_argument("--string", "-s", type=str, help="KV string")

    def __init__(self, args):
        self.key = args.key
        if args.value is not None:
            self.value = bytes.fromhex(args.value)
        elif args.string is not None:
            if self.key not in kv_defs.slots.ID_MAPPING:
                sys.exit(f"Key ID {self.key} not known, cannot validate")
            kv_type = kv_defs.slots.ID_MAPPING[self.key]
            # Validate key type is a string
            if (
                len(kv_type._fields_) != 0
                or not hasattr(kv_type, "vla_field")
                or not isinstance(kv_type.vla_field, tuple)
                or kv_type.vla_field[1] != kv_defs.structs.kv_string
            ):
                sys.exit(f"Key ID {self.key} is not a string value")

            str_val = args.string.encode("utf-8") + b"\x00"
            self.value = len(str_val).to_bytes(1, "little") + str_val
        else:
            raise NotImplementedError("Unimplmented value parsing")

    def request_struct(self):
        kv_struct = self.kv_store_value_factory(self.key, self.value)
        request_bytes = bytes(kv_struct)
        return bytes(self.request(1)) + request_bytes

    def handle_response(self, return_code, response):
        if return_code != 0:
            print(f"Invalid data buffer ({errno.strerror(-return_code)})")
            return

        def print_status(name, rc):
            if rc < 0:
                print(f"{name} failed to write ({os.strerror(-rc)})")
            elif rc == 0:
                print(f"{name} already matched")
            else:
                print(f"{name} updated")

        print_status(f"{self.key} value", response.rc[0])



from enum import Enum


class APIKeyResourcePerm(str, Enum):
    ADMIN = "admin"
    CREATE = "create"
    DELETE = "delete"
    UPDATE = "update"
    VIEW = "view"

    def __str__(self) -> str:
        return str(self.value)
